from ..utilities import JobStatus, JobResponse
from braket.devices import LocalSimulator
import strangeworks
from strangeworks.braket import get_backends, run_circuit, get_circuit_results


class BraketFaaS:
    def __init__(self, backendData: dict):
        self.token = backendData["providerToken"]
        self.swUser = backendData["backendInfo"].get("swUser")
        self.provider = backendData.get("provider")
        self.backendName = backendData["name"]

    def init_sw_provider(self):
        strangeworks.authenticate(username=self.swUser, api_key=self.token)

    def submit_job(self, qcircuit, shots):
        jobResult = {}
        if self.provider == "qfaas":
            backend = LocalSimulator()
            job = backend.run(qcircuit, shots=shots)
            jobResult = job.result()
            hub = "qfaas-internal"
        elif self.provider == "braket-sw":
            self.init_sw_provider()
            backend = self.backendName
            job = run_circuit(circuit=qcircuit, backend=backend, shots=shots)
            jobResult = get_circuit_results(job)
            hub = self.swUser
        if jobResult:
            jobStatus = JobStatus(
                "DONE",
                "Job is successfully executed on Local Simulator",
            )
            providerJobId = jobResult.task_metadata.id
        else:
            jobStatus = JobStatus("FAILED", "Job is failed")
            providerJobId = ""

        jobResponse = JobResponse(
            providerJobId=providerJobId,
            jobStatus=jobStatus,
            backend={
                "name": self.backendName,
                "hub": hub,
            },
            jobResult=dict(jobResult.measurement_counts),
        )

        return jobResponse



from typing import Optional, Union, List, Tuple
import array, math, fractions


class QFShor:
    """This class is inherited from Qiskit Shor Algorithm"""

    def __init__(self):
        self._factor = []
        self._total_counts = 0
        self._successful_counts = 0

    @property
    def factors(self) -> List[List[int]]:
        """returns factors"""
        return self._factors

    @factors.setter
    def factors(self, value: List[List[int]]) -> None:
        """set factors"""
        self._factors = value

    @property
    def total_counts(self) -> int:
        """returns total counts"""
        return self._total_counts

    @total_counts.setter
    def total_counts(self, value: int) -> None:
        """set total counts"""
        self._total_counts = value

    @property
    def successful_counts(self) -> int:
        """returns successful counts"""
        return self._successful_counts

    @successful_counts.setter
    def successful_counts(self, value: int) -> None:
        """set successful counts"""
        self._successful_counts = value

    def get_factors(self, N: int, a: int, measurement: str) -> Optional[List[int]]:
        """Inherited from Qiskit - Apply the continued fractions to find r and the gcd to find the desired factors."""
        x_final = int(measurement, 2)
        # logger.info("In decimal, x_final value for this result is: %s.", x_final)

        if x_final <= 0:
            fail_reason = "x_final value is <= 0, there are no continued fractions."
        else:
            fail_reason = None
            # logger.debug("Running continued fractions for this case.")

        # Calculate T and x/T
        T_upper = len(measurement)
        T = pow(2, T_upper)
        x_over_T = x_final / T

        # Cycle in which each iteration corresponds to putting one more term in the
        # calculation of the Continued Fraction (CF) of x/T

        # Initialize the first values according to CF rule
        i = 0
        b = array.array("i")
        t = array.array("f")

        b.append(math.floor(x_over_T))
        t.append(x_over_T - b[i])

        exponential = 0.0
        while i < N and fail_reason is None:
            # From the 2nd iteration onwards, calculate the new terms of the CF based
            # on the previous terms as the rule suggests
            if i > 0:
                b.append(math.floor(1 / t[i - 1]))
                t.append((1 / t[i - 1]) - b[i])  # type: ignore

            # Calculate the denominator of the CF using the known terms
            denominator = self.calculate_continued_fraction(b)

            # Increment i for next iteration
            i += 1

            if denominator % 2 == 1:
                # logger.debug("Odd denominator, will try next iteration of continued fractions.")
                continue

            # Denominator is even, try to get factors of N
            # Get the exponential a^(r/2)

            if denominator < 1000:
                exponential = pow(a, denominator / 2)

            # Check if the value is too big or not
            if exponential > 1000000000:
                fail_reason = "denominator of continued fraction is too big."
            else:
                # The value is not too big,
                # get the right values and do the proper gcd()
                putting_plus = int(exponential + 1)
                putting_minus = int(exponential - 1)
                one_factor = math.gcd(putting_plus, N)
                other_factor = math.gcd(putting_minus, N)

                # Check if the factors found are trivial factors or are the desired factors
                if any(factor in {1, N} for factor in (one_factor, other_factor)):
                    # logger.debug("Found just trivial factors, not good enough.")
                    # Check if the number has already been found,
                    # (use i - 1 because i was already incremented)
                    if t[i - 1] == 0:
                        fail_reason = (
                            "the continued fractions found exactly x_final/(2^(2n))."
                        )
                else:
                    # Successfully factorized N
                    return sorted((one_factor, other_factor))

        # Search for factors failed, write the reason for failure to the debug logs
        # print(
        #     "Cannot find factors from measurement %s because %s",
        #     measurement,
        #     fail_reason or "it took too many attempts.",
        # )
        return None

    def calculate_continued_fraction(self, b: array.array) -> int:
        """Calculate the continued fraction of x/T from the current terms of expansion b."""

        x_over_T = 0

        for i in reversed(range(len(b) - 1)):
            x_over_T = 1 / (b[i + 1] + x_over_T)

        x_over_T += b[0]

        # Get the denominator from the value obtained
        frac = fractions.Fraction(x_over_T).limit_denominator()

        # logger.debug("Approximation number %s of continued fractions:", len(b))
        # logger.debug("Numerator:%s \t\t Denominator: %s.", frac.numerator, frac.denominator)
        return frac.denominator



from ..utilities import JobStatus, JobResponse

import cirq


class CirqFaaS:
    def __init__(self, backendData: dict):
        self.token = backendData["providerToken"]
        self.swUser = backendData["backendInfo"].get("swUser")
        self.provider = backendData.get("provider")
        self.backendName = backendData["name"]

    def submit_job(self, qcircuit, shots):
        jobResult = {}
        if self.provider == "qfaas":
            simulator = cirq.Simulator()
            job = simulator.run(qcircuit, repetitions=shots)
            jobData = job.data
            # Get counts (export the result similar to Qiskit and Braket)
            jobResult = self.get_counts(jobData)
            hub = "qfaas-internal"

        if jobResult:
            jobStatus = JobStatus(
                "DONE",
                "Job is successfully executed on Local Simulator",
            )
            providerJobId = "QFaaS-Internal-Cirq-Simulation-Job"
        else:
            jobStatus = JobStatus("FAILED", "Job is failed")
            providerJobId = ""

        jobResponse = JobResponse(
            providerJobId=providerJobId,
            jobStatus=jobStatus,
            backend={
                "name": self.backendName,
                "hub": hub,
            },
            jobResult=jobResult,
        )
        return jobResponse

    def get_counts(self, jobData) -> dict:
        jobData["result"] = jobData.apply(lambda x: "".join(x.astype(str)), 1)
        counts = jobData.groupby(["result"]).size()
        countsDict = counts.to_dict()
        return countsDict



from enum import auto
import requests
from .dataUtils import RequestData
from .utilities import JobResponse, Utils, JobStatus


class Backend:
    def __init__(self, requestData: RequestData, circuit):
        self.sdk = requestData.sdk
        self.gateway = requestData.gateway
        self.headers = requestData.headers
        self.input = requestData.input
        self.rQubit = self.get_qubit_number(circuit)
        self.shots = requestData.shots
        self.wait = requestData.wait
        self.provider = requestData.provider
        self.backendRequest = self.generate_backend_request(requestData.data)
        # Get backend data from QFaaS Core API and generate backend instance
        self.backendData = self.get_backend_data()

    def get_qubit_number(self, circuit) -> int:
        if self.sdk == "qiskit":
            rQubit = int(circuit.num_qubits)
        elif self.sdk == "braket":
            rQubit = int(circuit.qubit_count)
        elif self.sdk == "cirq":
            rQubit = int(len(circuit.all_qubits()))
        elif self.sdk == "qsharp":
            rQubit = int(self.input)  # Temporarily
        return rQubit

    def generate_backend_request(self, jsonData):
        autoSelect = jsonData.get("autoSelect")
        backendName = "" if autoSelect else jsonData.get("backendName")
        backendType = (
            jsonData.get("backendType")
            if jsonData.get("backendType") in ["simulator", "qpu"]
            else "simulator"
        )

        backendRequest = {
            "sdk": self.sdk,
            "rQubit": self.rQubit,
            "provider": self.provider,
            "type": backendType,
            "autoSelect": autoSelect,
            "backendName": backendName,
        }

        return backendRequest

    def get_backend_data(self):
        response = requests.post(
            self.gateway + "/backend/select",
            json=self.backendRequest,
            headers=self.headers,
        )
        if response.json().get("code") == 200:
            try:
                backendData = response.json().get("data")[0]
            except:
                return None
            return backendData
        else:
            return None

    def submit_job(self, circuit, interval: int = 3, timeout: int = 30) -> JobResponse:
        jobResponse = JobResponse()
        shots = self.shots
        wait = self.wait
        if circuit and self.backendData and interval > 0 and interval <= timeout:
            if self.sdk == "qiskit":
                from .sdk.qiskit import QiskitFaaS

                beInstance = QiskitFaaS(self.backendData)
                jobResponse = beInstance.submit_job(
                    circuit, shots=shots, interval=interval, timeout=timeout, wait=wait
                )
            elif self.sdk == "braket":
                from .sdk.braket import BraketFaaS

                beInstance = BraketFaaS(self.backendData)
                jobResponse = beInstance.submit_job(circuit, shots=shots)

            elif self.sdk == "cirq":
                from .sdk.cirq import CirqFaaS

                beInstance = CirqFaaS(self.backendData)
                jobResponse = beInstance.submit_job(circuit, shots=shots)

            elif self.sdk == "qsharp":
                from .sdk.qsharp import QsharpFaaS

                beInstance = QsharpFaaS(self.backendData)
                jobResponse = beInstance.submit_job(circuit, self.input)
        elif self.backendData is None:
            jobResponse.backend = {"name": "Not found"}
            jobResponse.jobStatus.details = (
                "No backend matching the criteria. Please check your input."
            )
        return jobResponse

    def generate_job_response(self, job):
        jobResult = None
        jobStatus = self.monitor_job(job, 1, 2)
        if jobStatus.status == "DONE":
            jobResult = job.result()
        output = {
            "providerJobId": job.job_id(),
            "jobStatus": jobStatus,
            "backend": job.backend().name(),
            "jobResult": jobResult,
        }
        return output

    def get_job_results(self, job):
        if self.sdk == "qiskit":
            jobResult = job.result()
            return jobResult



class RequestData:
    
    def __init__(self, event, sdk: str, gateway: str = "http://qfaas-core.qfaas.svc.cluster.local/api"):
        self.data = event.json
        self.input = event.json.get("input")
        self.shots = event.json.get("shots")
        self.provider = event.json.get("provider")
        self.wait = bool(event.json.get("waitForResult"))
        self.postProcessOnly = bool(event.json.get("postProcessOnly"))
        self.headers = {"Authorization": event.headers.get("Authorization")}
        self.sdk = sdk
        self.gateway = gateway
        self.jobRawResult = ""
        if self.postProcessOnly:
            try:
                self.jobRawResult = event.json.get("jobRawResult")
            except:
                pass
    

    


from .algorithms.shor import QFShor


class JobStatus(object):
    def __init__(self, status="", details=""):
        self.status = status if status else "ERROR"
        self.details = (
            details
            if details
            else "Job is failed or incomplete. Please try again later."
        )


class JobResponse(object):
    def __init__(
        self,
        providerJobId: str = "",
        jobStatus: JobStatus = None,
        backend={},
        jobResult={},
    ):
        self.providerJobId = providerJobId if providerJobId else ""
        self.jobStatus = jobStatus if jobStatus else JobStatus()
        self.backend = backend if backend else {"name": "UNKNOWN"}
        self.jobResult = jobResult if jobResult else {}


class Utils:
    def __init__(self):
        pass

    def generate_response(jobResponse: JobResponse) -> dict:
        if jobResponse:
            statusCode = 201  # not yet finished
            if jobResponse.jobStatus.status == "DONE":
                statusCode = 200
            jobResponse.jobStatus = vars(jobResponse.jobStatus)
            jobResponse = vars(jobResponse)  # Object to directory
            response = {"statusCode": statusCode, "body": jobResponse}
        else:
            response = {
                "statusCode": 500,
                "body": "Error in function code. Please contact the developer.",
            }
        return response

    def generate_dict_output(data, details) -> dict:
        output = {"data": data, "details": details}
        return output

    def qrng_counts_post_process(job) -> JobResponse:
        # If input type = JobResponse or have jobRawResult
        jobRawResult = {}
        if type(job) is JobResponse:
            if job.jobResult:
                jobRawResult = job.jobResult
        elif hasattr(
            job, "jobRawResult"
        ):  # Type is FunctionInvocationSchema (post-process only)
            jobRawResult = job.jobRawResult
            job = JobResponse()
        else:
            job = JobResponse()
            return job

        if jobRawResult:
            job.jobResult = PostProcess.counts_qrng(jobRawResult)
            job.jobStatus = JobStatus("DONE", "Job post processing completed")
        return job

    # General Post processing selector (type="qrng", "shor")
    def qfaas_post_process(job, ptype, requestData) -> JobResponse:
        # If input type = JobResponse or have jobRawResult
        jobRawResult = {}
        if type(job) is JobResponse:
            if job.jobResult:
                jobRawResult = job.jobResult
        elif hasattr(
            job, "jobRawResult"
        ):  # Type is FunctionInvocationSchema (post-process only)
            jobRawResult = job.jobRawResult
            job = JobResponse()
        else:
            job = JobResponse()
            return job

        if jobRawResult:
            if ptype == "qrng":
                job.jobResult = PostProcess.counts_qrng(jobRawResult)
            elif ptype == "shor":
                # Get input number
                inputNumber = requestData.input
                aShor = 2  # Default a = 2
                job.jobResult = PostProcess.shor_postprocess(
                    jobRawResult, inputNumber, aShor=2  # Default a = 2
                )
            job.jobStatus = JobStatus("DONE", "Job post processing completed")
        return job


# Post-Processing functions
class PostProcess:
    def __init__(self):
        pass

    def counts_qrng(jobCounts: dict) -> dict:
        result = max(jobCounts, key=jobCounts.get)
        occurrence = max(jobCounts.values())
        allPossibleValues = {
            k: int(k, 2) for k, v in jobCounts.items() if v == occurrence
        }
        details = {
            "decimalValue": int(result, 2),
            "numberOfOccurence": occurrence,
            "allPossibleValues": allPossibleValues,
        }
        jobResult = {"data": result, "details": details}
        return jobResult

    def shor_postprocess(counts: dict, inputNumber: int, aShor: int = 2):
        qfshor = QFShor()
        for measurement in list(counts.keys()):
            # Get the x_final value from the final state qubits
            factors = qfshor.get_factors(inputNumber, aShor, measurement)

            if factors:
                # logger.info("Found factors %s from measurement %s.", factors, measurement)
                qfshor.successful_counts = qfshor.successful_counts + 1
                if factors not in qfshor.factors:
                    qfshor.factors.append(factors)
        result = qfshor.factors
        details = {"all_counts": counts, "successful_counts": qfshor.successful_counts}
        jobResult = {"data": result, "details": details}
        return jobResult



from ..utilities import JobStatus, JobResponse

import qsharp


class QsharpFaaS:
    def __init__(self, backendData: dict):
        self.token = backendData["providerToken"]
        self.swUser = backendData["backendInfo"].get("swUser")
        self.provider = backendData.get("provider")
        self.backendName = backendData["name"]

    def submit_job(self, qcircuit, input):
        jobResult = {}
        if self.provider == "qfaas":
            job = qcircuit.simulate(input=input)
            jobResult = {"data": job}
            hub = "qfaas-internal"

        if jobResult:
            jobStatus = JobStatus(
                "DONE",
                "Job is successfully executed on Local Simulator",
            )
            providerJobId = "QFaaS-Internal-QSharp-Simulation-Job"
        else:
            jobStatus = JobStatus("FAILED", "Job is failed")
            providerJobId = ""

        jobResponse = JobResponse(
            providerJobId=providerJobId,
            jobStatus=jobStatus,
            backend={
                "name": self.backendName,
                "hub": hub,
            },
            jobResult=jobResult,
        )
        return jobResponse



from qiskit import IBMQ, transpile, Aer
from ..utilities import JobStatus, JobResponse
import time

class QiskitFaaS:
    
    def __init__(self, backendData):
        self.token = backendData["providerToken"]
        self.hub = backendData["backendInfo"].get("hub")
        self.provider = backendData.get("provider")
        self.backendName = backendData["name"]
        self.backend = self.get_qiskit_backend()
    
    def init_ibmq_provider(self, token, hub):
        if IBMQ.active_account():
            if IBMQ.active_account()["token"] == token:
                provider = IBMQ.get_provider(hub=hub)
            else:
                IBMQ.disable_account()
                provider = IBMQ.enable_account(token, hub=hub)
        else:    
            provider = IBMQ.enable_account(token, hub=hub)
        return provider
    
    def get_qiskit_backend(self):
        if self.provider == "ibmq":
            ibmq = self.init_ibmq_provider(self.token, self.hub)
            backend = ibmq.get_backend(self.backendName)
        elif self.provider == "qfaas":
            backend = Aer.get_backend(self.backendName)
        return backend
    
    def transpile_circuit(self, qcircuit):
        transqc = transpile(qcircuit,self.backend)
        return transqc
        
    def run_job(self, qcircuit, shots):
        # Transpile circuit to adapt with backend basis gate
        qc = self.transpile_circuit(qcircuit)
        job = self.backend.run(qc, shots=shots)
        return job
    
    def submit_job(self, qcircuit, shots, interval, timeout, wait):
        if self.provider == "qfaas":
            try:
                job = self.run_job(qcircuit, shots)
                jobResult = dict(job.result().get_counts())
                jobStatus = JobStatus(
                    "DONE", "Job is successfully executed on Local Simulator"
                )
            except Exception as e:
                jobResult = {}
                jobStatus = JobStatus("ERROR", str(e))
            jobResponse = JobResponse(
                providerJobId="QFaaS-Internal-Qiskit-Simulation-Job",
                jobStatus=jobStatus,
                backend={
                    "name": self.backendName,
                    "hub": "qfaas-internal",
                },
                jobResult=jobResult,
            )
        elif self.provider == "ibmq":
            job = self.run_job(qcircuit, shots)
            jobResult = {}
            # wait for result processing
            if wait:
                # Track the status of job up to timeout (60s by default)
                jobStatus = self.ibmq_job_monitor(job, interval, timeout)
            else:
                # Track the status of job up to 10 seconds
                jobStatus = self.ibmq_job_monitor(job, 2, 5)

            if jobStatus.status == "DONE":
                counts = job.result()
                jobResult = dict(counts.get_counts())

            jobResponse = JobResponse(
                providerJobId=job.job_id(),
                jobStatus=jobStatus,
                backend={
                    "name": job.backend().name(),
                    "hub": job.backend().hub,
                },
                jobResult=jobResult,
            )
        return jobResponse
            
            
    
    def ibmq_job_monitor(self, job, interval : int, timeout: int):
        """Monitor job status at IBM Quantum

        Args:
        - job (IBMQJob): Job instance
        - interval (int): Interval time to check job status (in seconds)

        Returns:
        - Job status
        """
        status = job.status()
        proccessingTime = 0
        
        while status.name not in ["DONE", "CANCELLED", "ERROR"]:
            time.sleep(interval)
            status = job.status()
            msg = status.value
            if status.name == "QUEUED":
                details = msg + " (%s)" % job.queue_position()
            proccessingTime += interval
            if proccessingTime >= timeout:
                break
        msg = status.value
        details = msg
        jobStatus = JobStatus(status.name, details)
        return jobStatus
        